use crate::{primitives::Box3, stl::parser::StlSolid};

/// A node representing a node single in the bounding volume hierarchy
#[derive(Default)]
struct BvhNode {
    aabb: Box3,
    left_child: usize,
    right_child: usize,
    prim_idx: usize,
    prim_count: usize,
}

/// Struct representing the total bounding volume hierarchy
pub struct Bvh {
    nodes: Vec<BvhNode>,
    tri_idx: Vec<usize>,
}

/// Return type for get_subdivision_position_and_axis
struct SplitPlane {
    axis: usize,
    pos: f64,
}

impl Bvh {
    /// Create a new Bounding volume hierarchy from an STL solid.
    pub fn new(solid: &StlSolid) -> Self {
        // Initialize the root node
        let mut nodes = vec![BvhNode {
            prim_idx: 0,
            prim_count: solid.triangles.len(),
            ..Default::default()
        }];
        nodes[0]
            .aabb
            .shrink_wrap_primitives(&solid.vertices, &solid.triangles);

        let mut bvh = Bvh {
            nodes,
            tri_idx: Vec::from_iter(0..solid.triangles.len()),
        };

        bvh.subdivide(0, solid);
        bvh
    }

    /// Subdivide a node, subdivision doesn't occur if the node has only 2 primitives
    pub fn subdivide(&mut self, node_idx: usize, solid: &StlSolid) {
        let SplitPlane {
            axis: split_axis,
            pos: split_position,
        } = self.get_split_plane(node_idx);

        let prim_count = self.nodes[node_idx].prim_count;
        let prim_idx = self.nodes[node_idx].prim_idx;
        // Reorganize the triangles so the triangles are ordered, this allows triangles belonging
        // to a specific node to be adjacent
        let mut start_idx = prim_idx;
        let mut end_idx = prim_idx + prim_count - 1;
        while start_idx <= end_idx {
            if solid.triangle_centroids[self.tri_idx[start_idx]][split_axis] < split_position as f32
            {
                start_idx += 1;
            } else {
                self.tri_idx.swap(start_idx, end_idx);
                end_idx -= 1;
            }
        }

        // Now split the primitives into two nodes
        let left_count = start_idx - prim_idx;
        let right_count = prim_count - left_count;

        // If all primitives belong to either the left or right half exit because the parent node
        // is a leaf
        if left_count == 0 || left_count == prim_count {
            return;
        }

        // Otherwise split the nodes into left and right child
        let left_child = self.nodes.len();
        let right_child = self.nodes.len() + 1;

        // FIX: Having to copy the [usize; 3] that defines each triangle just to create the
        // axis-aligned bounding box seems wasteful although it makes the code interface a lot
        // cleaner. Maybe spend some time to redesign this part to make it cleaner once we really
        // start focusing on performance tuning
        let left_prim_triangles: Vec<_> = (prim_idx..left_count)
            .map(|idx| solid.triangles[idx])
            .collect();
        let mut aabb = Box3::default();
        aabb.shrink_wrap_primitives(&solid.vertices, &left_prim_triangles);

        self.nodes.push(BvhNode {
            aabb,
            left_child: 0,
            right_child: 0,
            prim_idx,
            prim_count: left_count,
        });

        let right_prim_triangles: Vec<_> = ((prim_idx + left_count)..prim_count)
            .map(|idx| solid.triangles[idx])
            .collect();
        let mut aabb = Box3::default();
        aabb.shrink_wrap_primitives(&solid.vertices, &right_prim_triangles);

        self.nodes.push(BvhNode {
            aabb,
            left_child: 0,
            right_child: 0,
            prim_idx: prim_idx + left_count,
            prim_count: right_count,
        });
        self.nodes[node_idx].prim_count = 0;
        self.nodes[node_idx].left_child = left_child;
        self.nodes[node_idx].right_child = right_child;

        // Recurse to divide children
        self.subdivide(left_child, solid);
        self.subdivide(right_child, solid);
    }

    /// Calculate the optimal split plane axis and position.
    /// This function should use the surface area heuristic to select the optimal split
    fn get_split_plane(&self, node_idx: usize) -> SplitPlane {
        // Initial approach just splits halfway along the longest axis
        // TODO: Improve this to use the surface area heuristic
        let axis = (0..3).fold(0, |acc, a| {
            if self.nodes[node_idx].aabb.extents[acc] < self.nodes[node_idx].aabb.extents[a] {
                a
            } else {
                acc
            }
        });

        let pos = self.nodes[node_idx].aabb.extents[axis] / 2.0;
        SplitPlane { axis, pos }
    }
}
